#include <stdlib.h>
#include <stdio.h>
#include "local.h"

#define MAX_DIM 10      // Maximum number of dimensions
#define MAX_BUFFER 256  // Max buffer size
#define NFILTERS   18
#define ISSTD      0    // Standard/Nonstandard wavelet basis (wvlt library)



extern int RecurFilter_(double *DATA, double *WORK, int* size,char* Wavelet,int* dim, int* iscoh, int verbose);

int main(int argc, char **argv)
{
int i, j, data_type, filter_alg, Ndim, Nelem, Nelem2, nf, coh, verbose;
int dimensions[MAX_DIM];
float *data4 = NULL;
double *data8 = NULL, *work = NULL, *work2 = NULL, *store = NULL;
FILE *fin, *fout1, *fout2;
waveletfilter *pftr;
char buffer[MAX_BUFFER], buffer2[MAX_BUFFER], filter[MAX_BUFFER]; 
char ext_dwt[20], ext_coh[20], ext_incoh[20];
char* filters_available[NFILTERS] = { "AdelsonSimoncelliHingorani",
				      "AntoniniBarlaudMatheiuDaubechies_4_4",     
				      "BattleLemarie",     
				      "BurtAdelson",    
				      "Coifman_12", // (Coifman wavelet with 12 coefficients i.e. 4 vanishing moments) 
				      "Daubechies_4",
				      "Daubechies_6",
				      "Daubechies_8",     
				      "Daubechies_10",     
				      "Daubechies_12",
				      "Daubechies_20",     
				      "Haar",     
				      "Pseudocoiflet_4_4",     
				      "Spline_2_2",     
				      "Spline_2_4",
				      "Spline_3_3",
				      "Spline_3_5",
				      "Spline_3_7" };


 if ( argc != 2 )
 {
   perror("  Input file missing !\n\n");
   exit(EXIT_FAILURE);    
 }

 for(i=0;i<MAX_DIM;i++)
   dimensions[i] = 1;

// Reading the parameter file;
// Verbosity level;
  fscanf(stdin,"%d", &verbose);
  fgets(buffer, MAX_BUFFER, stdin);

// Dimension of the field;
  fscanf(stdin,"%d", &Ndim);
  fgets(buffer, MAX_BUFFER, stdin);
  if ( Ndim < 3 )
  {
    if ( verbose )
      fprintf(stderr, "  Dimension= %d \n", Ndim);
  }
  else
  {
    perror("  Dimension must be less than 4 !\n\n");
    exit(EXIT_FAILURE);    
  }

// Size of the field;
  Nelem = 1;
  Nelem2 = 1;
  if ( verbose )  
    fprintf(stderr, "  Size:  ");
  for(i=0;i<Ndim;i++)
    if ( fscanf(stdin, "%d", &dimensions[i]) != 1 )
    {
      perror("  Problems reading dimensions !\n\n");
      exit(EXIT_FAILURE);    
    }
    else
    {
      if ( verbose )
	fprintf(stderr," %d ", dimensions[i]);
      Nelem  *= dimensions[i];
      Nelem2 *= dimensions[i] / 2;
    }
  fgets(buffer, MAX_BUFFER, stdin);
  if ( verbose )
    fprintf(stderr,"\n");

// Bytes per number;
  fscanf(stdin,"%d", &data_type);
  switch (data_type)
  {
    case 0:
      if ( verbose )
	fprintf(stderr,"  Reading ASCII numbers ... \n");
      break;      
    case 4:
      if ( verbose )	  
	fprintf(stderr,"  Reading floats ... \n");
      break;
    case 8:
      if ( verbose )
	fprintf(stderr,"  Reading doubles ... \n");
      break;   
    default:
      perror("  Wrong data tyle (size) !\n\n");
      exit(EXIT_FAILURE);    
      break;
  }
  fgets(buffer, MAX_BUFFER, stdin);


// Determining the filter;
  fscanf(stdin,"%s", &filter);
  i = 0;
  while ( i < NFILTERS && strcmp( filter, filters_available[i] ) != 0 ) i++;
  if ( i < NFILTERS )
  {
    if ( verbose )
      fprintf(stderr,"  Filter used: %s [%d] \n", filter, i);
  }
  else
  {
    perror("  Invalid filter !\n\n");
    exit(EXIT_FAILURE);    
  }
  fgets(buffer, MAX_BUFFER, stdin);

// Determining the filtering algorithm;
  fscanf(stdin,"%d", &filter_alg);
  switch (filter_alg)
  {
    case 0:
      if ( verbose )
	fprintf(stderr,"  Performing DWT only \n");
      break;      
    case 1:
      if ( verbose )
	fprintf(stderr,"  Using MAD filtering algorithm \n");
      break;      
    case 2:
      if ( verbose )
	fprintf(stderr,"  Using the recursive filtering algorithm \n");
      break;
    default:
      perror("  Invalid filtering algorithm !\n\n");
      exit(EXIT_FAILURE);    
      break;
  }
  fgets(buffer, MAX_BUFFER, stdin);

// Bytes per number;
  fscanf(stdin,"%s%s%s", ext_dwt, ext_coh, ext_incoh );
  if ( verbose )
    fprintf(stderr,"  Using the extensions: %s %s %s \n", ext_dwt, ext_coh, ext_incoh );
  fgets(buffer, MAX_BUFFER, stdin);


// Allocating the memory;
  work = malloc(Nelem*sizeof(double));
  data8 = malloc(Nelem*sizeof(double));

// for floats will need an additional array;
  if ( data_type == 4 )
    data4 = malloc(Nelem*data_type);
  if ( filter_alg == 1 )
    work2 = malloc(Nelem*sizeof(double));
    
// additional array will be needed for filtering;
  if ( filter_alg > 0 )
    store = malloc(Nelem*sizeof(double));
    
// Reading the input data
  switch (data_type)
  {
    case 0:
// ASCII input;
      fin = fopen( argv[1], "rt");     
      i = 0;
      while( EOF != fscanf(fin,"%lf", &data8[i++] ) );      
      if ( Nelem != (i - 1) )
      {
	perror("  Incorrect data input ! \n\n");
	exit(EXIT_FAILURE);
      }
      break;
    case 4:
      fin = fopen( argv[1], "rb");     
      if ( Nelem != fread( data4, 4, Nelem, fin) )
      {
	perror("  Incorrect data input ! \n\n");
	exit(EXIT_FAILURE);
      }
// Casting on double
      for(i=0;i<Nelem;i++)
	data8[i] = data4[i];
      break;

    case 8:
      fin = fopen( argv[1], "rb");  
      if ( Nelem != fread( data8, 8, Nelem, fin) )
      {
	perror("  Incorrect data input ! \n\n");
	exit(EXIT_FAILURE);
      }
      break;
  }
  fclose(fin);

// Performing the filtering;
  switch(filter_alg)
  {
    case 0:  
      if ( (pftr = wfltr_select(filter, &nf)) == NULL )
      {
	fprintf(stderr,"  Error selecting filter: %d (%s) !\n\n", nf, filter);
	exit(nf);
      }      
      wxfrm_dand( data8, dimensions, Ndim, 1, ISSTD, pftr, work );
      break;
    case 1:
// This case is unfortunately incomplete;
// storing the original field;
      for(i=0;i<Nelem;i++) 
	store[i] = data8[i];       
      coh = 1; 
      MedianWltFilter_( data8, work, work2, dimensions, filter, &Ndim, &coh, &Nelem2); 
// Calculating the incoherent field;
      for(i=0;i<Nelem;i++)
	work[i] = store[i] - data8[i];
      break;
    case 2:
      coh = 1;
// storing the original field;
      for(i=0;i<Nelem;i++)
	store[i] = data8[i];      
      RecurFilter_(data8, work, dimensions, filter, &Ndim, &coh, verbose);
// Calculating the incoherent field;
      for(i=0;i<Nelem;i++)
	work[i] = store[i] - data8[i];
      break;
    }

// Saving data;  
  if ( filter_alg == 0 )
  {
    sprintf(buffer,"%s.%s", argv[1], ext_dwt);
    if ( verbose )
      fprintf(stderr,"  Saving file %s \n", buffer);
    if ( data_type == 0 )
    {
// ASCII output;
      fout1 = fopen(buffer,"wt");
// for the moment it works only in 1D and 2D;	
      for(i=0;i<dimensions[0];i++)
      {
	for(j=0;j<dimensions[1];j++)
	  fprintf(fout1," % 14lf", work[i*dimensions[1]+j]);
	fprintf(fout1,"\n");
      }
    }
    else
    {
 // binary (double) output;
      fout1 = fopen(buffer,"wb");
      fwrite( work, sizeof(double), Nelem, fout1);      
    }
    fclose(fout1);
  }
  else
  {
    sprintf(buffer ,"%s.%s", argv[1], ext_coh);
    sprintf(buffer2,"%s.%s", argv[1], ext_incoh);
    if ( verbose )
      fprintf(stderr,"  Saving files %s %s \n", buffer, buffer2);
    if ( data_type == 0 )
    {
// ASCII output;
      fout1 = fopen(buffer ,"wt");
      fout2 = fopen(buffer2,"wt");
// for the moment it works only in 1D and 2D;
      for(i=0;i<dimensions[0];i++)
      {
	for(j=0;j<dimensions[1];j++)
	{
	  fprintf(fout1," % 14lf", data8[i*dimensions[1]+j]);
	  fprintf(fout2," % 14lf",  work[i*dimensions[1]+j]);
	}
	fprintf(fout1,"\n");
	fprintf(fout2,"\n");
      }
    }
    else
    {
// binary (double) output;
      fout1 = fopen(buffer ,"wb");
      fout2 = fopen(buffer2,"wb");
      fwrite(data8, sizeof(double), Nelem, fout1);      
      fwrite( work, sizeof(double), Nelem, fout2);            
    }
    fclose(fout1); 
    fclose(fout2);
  }
  
  exit(EXIT_SUCCESS);
}

